"""
trepr package.

Package for reproducible and traceable analysis of trepr data.

Available modules
-----------------
:mod:`trepr.dataset`
    Organise dataset, consisting of data and metadata, and the metadata.
:mod:`trepr.io`
    Import trepr raw data in different formats including its metadata.
:mod:`trepr.processing`
    Process data in datasets.
:mod:`trepr.analysis`
    Analyse data in datasets.
:mod:`trepr.plotting`
    Plot data in datasets.
:mod:`trepr.report`
    Create reports with given datasets.

Utilities
---------
:mod:`trepr.exceptions`
    Exceptions for the trepr package


"""
